@@ -31,15 +31,22 @@ See https://bugs.launchpad.net/bugs/928486
 from itertools import count
 import linecache
 import sys
+import os
+import tokenize
 from types import ModuleType as module
 
+
 class GeneratedCodeLoader(object):
     """
     Object for use as a module's __loader__, to display generated
     source.
     """
     def __init__(self, source):
-        self.source = source.encode('utf8')
+        if sys.version_info >= (3,):
+            self.source = source
+        else:
+            self.source = source.encode('utf8')
+
     def get_source(self, name):
         return self.source
 
