@@ -15,6 +15,12 @@
 
 """Tests for the pybars compiler."""
 
+try:
+    str_class = unicode
+except NameError:
+    # Python 3 support
+    str_class = str
+
 from testtools import TestCase
 
 from pybars import Compiler
@@ -30,8 +36,8 @@ def render(source, context, helpers=None, partials=None, knownHelpers=None,
         real_partials = None
     else:
         real_partials = dict((key, compiler.compile(value))
-            for key, value in partials.items())
-    return unicode(template(context, helpers=helpers, partials=real_partials))
+            for key, value in list(partials.items()))
+    return str_class(template(context, helpers=helpers, partials=real_partials))
 
 
 class TestCompiler(TestCase):
@@ -62,7 +68,7 @@ class TestCompiler(TestCase):
         self.assertEqual(
             u"<ul><li>Yehuda Katz</li><li>Carl Lerche</li>"\
             "<li>Alan Johnson</li></ul>",
-            unicode(rendered))
+            str_class(rendered))
 
     def test_escapes(self):
         self.assertEqual(u"""\
