@@ -23,6 +23,14 @@
 from testtools import TestCase
 from testtools.matchers import Equals
 
+try:
+    str_class = unicode
+except NameError:
+    # Python 3 support
+    str_class = str
+
+import sys
+
 import pybars
 from pybars import (
     Compiler,
@@ -71,6 +79,14 @@ class TestAcceptance(TestCase):
         self.assertEqual("num: 0",
             render(u"num: {{num1/num2}}", {'num1': {'num2': 0}}))
 
+    def test_negative_int_literal(self):
+        self.assertEqual("str int int",
+            render(u"{{type \"string\"}} {{type 1}} {{type -1}}", None,
+            helpers={'type': lambda s, v: type(v).__name__}))
+        self.assertEqual("string 1 -1",
+            render(u"{{echo \"string\"}} {{echo 1}} {{echo -1}}", None,
+            helpers={'echo': lambda s, v: str(v)}))
+
     def test_newlines(self):
         self.assertThat(u"Alan's\nTest", RendersItself())
         self.assertThat(u"Alan's\rTest", RendersItself())
@@ -210,6 +226,14 @@ class TestAcceptance(TestCase):
             "goodbye cruel Alan! Goodbye cruel Alan! GOODBYE cruel Alan! ",
             render(source, context))
 
+    def test_parent_lookup(self):
+        source = u"{{#goodbyes}}{{text}} cruel {{@_parent.name}}! {{/goodbyes}}"
+        context = {'name': "Alan", 'goodbyes': [
+            {'text': "goodbye"}, {'text': "Goodbye"}, {'text': "GOODBYE"}]}
+        self.assertEqual(
+            "goodbye cruel Alan! Goodbye cruel Alan! GOODBYE cruel Alan! ",
+            render(source, context))
+
     def test_helper_with_complex_lookup(self):
         template = u"{{#goodbyes}}{{{link ../prefix}}}{{/goodbyes}}"
         context = {
@@ -244,7 +268,7 @@ class TestAcceptance(TestCase):
         context = {
             'prefix': '/root', 'goodbyes': [{'text': "Goodbye", 'url': "goodbye"}]}
         def link(this, options, prefix):
-            return u"<a href='" + unicode(prefix) + u"/" + this['url'] + u"'>" + unicode(options['fn'](this)) + u"</a>"
+            return u"<a href='" + str_class(prefix) + u"/" + this['url'] + u"'>" + str_class(options['fn'](this)) + u"</a>"
         self.assertEqual(u"<a href='/root/goodbye'>Goodbye</a>",
             render(template, context, helpers={'link': link}))
 
@@ -254,6 +278,12 @@ class TestAcceptance(TestCase):
         context = {'omg': "OMG!", 'outer': [{'inner': [{'text': "goodbye"}]}]}
         self.assertEqual(u"Goodbye cruel OMG!", render(template, context))
 
+    def test_root_lookup(self):
+        template = u"{{#outer}}Goodbye "\
+            u"{{#inner}}cruel {{@root.top}}{{/inner}}{{/outer}}"
+        context = {'top': "world", 'outer': [{'inner': [{'text': "goodbye"}]}]}
+        self.assertEqual(u"Goodbye cruel world", render(template, context))
+
     def test_block_helper(self):
         template = u"{{#goodbyes}}{{text}}! {{/goodbyes}}cruel {{world}}!"
         self.assertEqual(
@@ -271,7 +301,7 @@ class TestAcceptance(TestCase):
         template = u"<ul>{{#people}}<li>"\
             u"{{#link}}{{name}}{{/link}}</li>{{/people}}</ul>"
         def link(this, options):
-            return strlist(('<a href="/people/', unicode(this['id']), '">', options['fn'](this), '</a>'))
+            return strlist(('<a href="/people/', str_class(this['id']), '">', options['fn'](this), '</a>'))
         context = {"people": [
             {"name": "Alan", "id": 1},
             {"name": "Yehuda", "id": 2}
@@ -290,14 +320,14 @@ class TestAcceptance(TestCase):
         context = {'yehuda': {'name': "Yehuda"}}
         expected = u"<form><p>Yehuda</p></form>"
         def form(this, options, context):
-            return "<form>" + unicode(options['fn'](context)) + '</form>'
+            return "<form>" + str_class(options['fn'](context)) + '</form>'
         helpers = {'form': form}
         self.assertEqual(expected, render(template, context, helpers=helpers))
 
     def test_block_helper_passing_a_complex_path_context(self):
         source = u"{{#form yehuda/cat}}<p>{{name}}</p>{{/form}}"
         def form(this, options, context):
-            return u"<form>" + unicode(options['fn'](context)) + u"</form>"
+            return u"<form>" + str_class(options['fn'](context)) + u"</form>"
         context = {'yehuda': {'name': "Yehuda", 'cat': {'name': "Harold"}}}
         self.assertEqual("<form><p>Harold</p></form>",
             render(source, context, helpers={'form': form}))
@@ -308,9 +338,9 @@ class TestAcceptance(TestCase):
         def link(this, options):
             return (
                 "<a href='" + this['name'] + "'>" +
-                unicode(options['fn'](this)) + "</a>")
+                str_class(options['fn'](this)) + "</a>")
         def form(this, options, context):
-            return "<form>" + unicode(options['fn'](context)) + "</form>"
+            return "<form>" + str_class(options['fn'](context)) + "</form>"
         self.assertEqual(
             "<form><p>Yehuda</p><a href='Yehuda'>Hello</a></form>",
             render(source, {'yehuda': {'name': "Yehuda"}}, helpers={
@@ -332,12 +362,12 @@ class TestAcceptance(TestCase):
                 out = "<ul>"
                 for thing in context:
                     out += "<li>"
-                    out += unicode(options['fn'](thing))
+                    out += str_class(options['fn'](thing))
                     out += "</li>"
                 out += "</ul>"
                 return out
             else:
-                return "<p>" + unicode(options['inverse'](this)) + "</p>"
+                return "<p>" + str_class(options['inverse'](this)) + "</p>"
         context = {'people': [{'name': "Alan"}, {'name': "Yehuda"}]}
         empty = {'people': []}
         rootMessage = {'people': [], 'message': "Nobody's here"}
@@ -409,6 +439,28 @@ class TestAcceptance(TestCase):
         self.assertEqual("Dudes: Yehuda (http://yehuda) Alan (http://alan) ",
             render(source, context, partials={'dude': partial}))
 
+    def test_partials_too_many_args(self):
+        source = u'Dudes: {{>dude dudes "extra"}}'
+        partial = u"{{#this}}{{name}} ({{url}}) {{/this}}"
+        context = {
+            'dudes': [
+                {'name': "Yehuda", 'url': "http://yehuda"},
+                {'name': "Alan", 'url': "http://alan"}
+                ]}
+        self.assertRaises(Exception, render, source, context,
+            partials={'dude': partial})
+
+    def test_partials_kwargs(self):
+        source = u'Dudes: {{#dudes}}{{>dude url="http://example"}}{{/dudes}}'
+        partial = u"{{name}} ({{url}}) "
+        context = {
+            'dudes': [
+                {'name': "Yehuda", 'url': "http://yehuda"},
+                {'name': "Alan", 'url': "http://alan"}
+                ]}
+        self.assertEqual("Dudes: Yehuda (http://example) Alan (http://example) ",
+            render(source, context, partials={'dude': partial}))
+
     def test_partial_in_a_partial(self):
         source = u"Dudes: {{#dudes}}{{>dude}}{{/dudes}}"
         dude_src = u"{{name}} {{> url}} "
@@ -427,6 +479,22 @@ class TestAcceptance(TestCase):
     def test_rendering_undefined_partial_throws_an_exception(self):
         self.assertRaises(Exception, render, u"{{> whatever}}", {})
 
+    def test_root_nested_partial(self):
+        source = u"Dudes: {{#dudes}}{{>dude}}{{/dudes}}"
+        dude_src = u"{{name}} {{> url}} "
+        url_src = u"<a href='{{url}}' target='{{@root.target}}'>{{url}}</a>"
+        partials = {'dude': dude_src, 'url': url_src}
+        context = {
+            'target': '_blank',
+            'dudes': [
+                {'name': "Yehuda", 'url': "http://yehuda"},
+                {'name': "Alan", 'url': "http://alan"}
+                ]}
+        self.assertEqual(
+            "Dudes: Yehuda <a href='http://yehuda' target='_blank'>http://yehuda</a>"
+            " Alan <a href='http://alan' target='_blank'>http://alan</a> ",
+            render(source, context, partials=partials))
+
     def test_GH_14_a_partial_preceding_a_selector(self):
         source = u"Dudes: {{>dude}} {{another_dude}}"
         dude_src = u"{{name}}"
@@ -447,8 +515,8 @@ class TestAcceptance(TestCase):
             self.assertEqual(True, bool1)
             self.assertEqual(False, bool2)
             self.assertEqual(12, times)
-            return ("Hello " + param + " " + unicode(times) + " times: " +
-                unicode(bool1) + " " + unicode(bool2))
+            return ("Hello " + param + " " + str_class(times) + " times: " +
+                str_class(bool1) + " " + str_class(bool2))
         helpers = dict(hello=hello)
         self.assertEqual("Message: Hello world 12 times: True False",
             render(source, {}, helpers=helpers))
@@ -495,7 +563,7 @@ class TestAcceptance(TestCase):
         reference = "testing 1, 2, 3"
         instance = strlist([reference])
         self.assertIsInstance(instance, strlist)
-        self.assertEqual(unicode(reference), unicode(instance))
+        self.assertEqual(str_class(reference), str_class(instance))
 
     def test_if_a_context_is_not_found_helperMissing_is_used(self):
         def link_to(this, helpername, context):
@@ -514,9 +582,9 @@ class TestAcceptance(TestCase):
 
     def test_Unknown_helper_in_knownHelpers_only_mode_should_be_passed_as_undefined(self):
         source = u"{{{typeof hello}}}"
-        self.assertEqual("<type 'NoneType'>",
+        self.assertEqual("<type 'NoneType'>" if sys.version_info < (3,) else "<class 'NoneType'>",
             render(source, {}, helpers=dict(
-                typeof=lambda this, arg: unicode(type(arg)), hello=lambda this: "foo"),
+                typeof=lambda this, arg: str_class(type(arg)), hello=lambda this: "foo"),
             knownHelpers=set(['typeof']), knownHelpersOnly=True))
 
     def test_Builtin_helpers_available_in_knownHelpers_only_mode(self):
@@ -592,6 +660,42 @@ class TestAcceptance(TestCase):
         self.assertEqual(u"cruel world!",
             render(source, {'goodbye': lambda this: None, 'world': "world"}))
 
+    def test_resolve_with_attrs(self):
+        class TestAttr():
+            @property
+            def text(self):
+                return 'Hello'
+
+        class TestGet():
+            def get(self, name):
+                return {'text': 'Hi'}.get(name)
+
+        source = u"{{#each .}}{{test.text}}! {{/each}}"
+        context = [
+            {'test': TestAttr()},
+            {'test': TestGet()},
+            {'test': {'text': 'Goodbye'}}
+        ]
+        self.assertEqual("Hello! Hi! Goodbye! ",
+            render(source, context))
+
+    def test_list_context(self):
+        source = u"{{#each .}}{{#each .}}{{text}}! {{/each}}cruel world!{{/each}}"
+        context = [[{'text': "goodbye"}, {'text': "Goodbye"}, {'text': "GOODBYE"}]]
+        self.assertEqual("goodbye! Goodbye! GOODBYE! cruel world!",
+            render(source, context))
+
+    def test_context_with_attrs(self):
+        class TestContext():
+            @property
+            def text(self):
+                return 'Goodbye'
+
+        source = u"{{#each .}}{{text}}! {{/each}}cruel world!"
+        context = [TestContext()]
+        self.assertEqual("Goodbye! cruel world!",
+            render(source, context))
+
     def test_each(self):
         source = u"{{#each goodbyes}}{{text}}! {{/each}}cruel {{world}}!"
         context = {'goodbyes':
@@ -602,6 +706,83 @@ class TestAcceptance(TestCase):
         self.assertEqual("cruel world!",
             render(source, {'goodbyes': [], 'world': "world"}))
 
+    def test_each_this(self):
+        source = u"{{#each name}}{{capitalize this}} {{/each}}"
+        helpers = {'capitalize': lambda this, value: value.upper()}
+        context = {
+            'name': ['John', 'James']
+        }
+        self.assertEqual("JOHN JAMES ",
+            render(source, context, helpers=helpers))
+
+    def test_each_of_None(self):
+        self.assertEqual(u"Goodbye cruel world!",
+            render(u"Goodbye {{^each things}}cruel{{/each}} world!",
+                {'things': None}))
+
+    def test_each_of_empty_list(self):
+        self.assertEqual(u"Goodbye cruel world!",
+            render(u"Goodbye {{#each things}}happy {{^}}cruel {{/each}}world!",
+                {'things': []}))
+
+    def test_each_of_truthy_non_iterable_object(self):
+        self.assertEqual(u"Goodbye cruel world!",
+            render(u"Goodbye {{#each things}}happy {{^}}cruel {{/each}}world!",
+                {'things': True}))
+
+    def test_each_with_object_and_key(self):
+        source = u"{{#each goodbyes}}{{@key}}. {{text}}! {{/each}}cruel {{world}}!"
+        context = {
+            'goodbyes': {
+                "<b>#1</b>": {'text': "goodbye"},
+                2: {'text': "GOODBYE"}
+            }, 'world': "world"};
+        self.assertIn(
+            render(source, context),
+
+            # Depending on iteration order, one will come before the other.
+            (
+                "&lt;b&gt;#1&lt;/b&gt;. goodbye! 2. GOODBYE! cruel world!",
+                "2. GOODBYE! &lt;b&gt;#1&lt;/b&gt;. goodbye! cruel world!"
+            )
+        )
+
+    def test_each_with_index(self):
+        source = u"{{#each goodbyes}}{{@index}}. {{text}}! {{/each}}cruel {{world}}!"
+        context = {
+            'goodbyes': [{'text': "goodbye"}, {'text': "Goodbye"}, {'text': "GOODBYE"}],
+            'world': "world"}
+        self.assertEqual(
+            "0. goodbye! 1. Goodbye! 2. GOODBYE! cruel world!",
+            render(source, context))
+
+    def test_each_with_nested_index(self):
+        source = u"{{#each goodbyes}}{{@index}}. {{text}}! {{#each ../goodbyes}}{{@index}} {{/each}}After {{@index}} {{/each}}{{@index}}cruel {{world}}!"
+        context = {
+            'goodbyes': [{'text': "goodbye"}, {'text': "Goodbye"}, {'text': "GOODBYE"}],
+            'world': "world"}
+        self.assertEqual(
+            "0. goodbye! 0 1 2 After 0 1. Goodbye! 0 1 2 After 1 2. GOODBYE! 0 1 2 After 2 cruel world!",
+            render(source, context))
+
+    def test_each_with_parent_index(self):
+        source = u"{{#each people}}{{#each foods}}{{../name}}({{@../index}}) likes {{name}}({{@index}}), {{/each}}{{/each}}"
+        context = {
+            'people': [
+                {
+                    'name': 'John',
+                    'foods': [{'name': 'apples'}, {'name': 'pears'}]
+                },
+                {
+                    'name': 'Jane',
+                    'foods': [{'name': 'grapes'}, {'name': 'pineapple'}]
+                }
+            ],
+        }
+        self.assertEqual(
+            "John(0) likes apples(0), John(0) likes pears(1), Jane(1) likes grapes(0), Jane(1) likes pineapple(1), ",
+            render(source, context))
+
     def test_log(self):
         source = u"{{log blah}}"
         context = {'blah': "whee"}
@@ -689,7 +870,7 @@ class TestAcceptance(TestCase):
     def test_block_helpers_can_take_an_optional_hash(self):
         source = u'{{#goodbye cruel="CRUEL" times=12}}world{{/goodbye}}'
         def goodbye(this, options, times, cruel):
-            return "GOODBYE " + cruel + " " + unicode(options['fn'](this)) + " " + str(times) + " TIMES"
+            return "GOODBYE " + cruel + " " + str_class(options['fn'](this)) + " " + str(times) + " TIMES"
         helpers = {'goodbye': goodbye}
         self.assertEqual("GOODBYE CRUEL world 12 TIMES",
             render(source, {}, helpers=helpers))
@@ -697,7 +878,7 @@ class TestAcceptance(TestCase):
     def test_block_helpers_can_take_an_optional_hash_with_booleans(self):
         def goodbye(this, options, cruel, _print):
             if _print is True:
-                return "GOODBYE " + cruel + " " + unicode(options['fn'](this))
+                return "GOODBYE " + cruel + " " + str_class(options['fn'](this))
             elif _print is False:
                 return "NOT PRINTING"
             else:
